/*

  SpringProx API
  --------------

  Copyright (c) 2000-2006 Pro-Active SAS - www.pro-active.fr

  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  History
  -------
  
  JDA 15/02/2003 : string list created from philips rc500/mv700 files
  JDA 16/05/2003 : added some strings
                   added helpers functions for Delphi and VB users
  LTC 26/02/2008 : added some strings for ISO 15693 and ICODE1

*/

#ifndef _BUILD_SPRINGPROX_EX_DLL

/**** SpringProxAPI/HostUtilities
 *
 * NAME
 *   SpringProxAPI :: Host utilities
 *
 **/
#include "sprox_api_i.h"

#ifndef SPROX_API_NO_MSG

/**f* SpringProxAPI/SPROX_GetErrorMessage
 *
 * NAME
 *   SPROX_GetErrorMessage
 *
 * DESCRIPTION
 *   Retrieve an english string explaining a function returned code
 *
 * INPUTS
 *   SWORD status  : SpringProxAPI function returned value
 *
 * RETURNS
 *   const TCHAR * : pointer to the translation string
 *
 * WARNING
 *   The string returned is physically static in the SpringProxAPI address
 *   space. Do not try to change it nor to free the returned pointer.
 *
 * VARIANTS
 *   SPROX_GetErrorMessageA : ASCII version
 *   SPROX_GetErrorMessageW : UNICODE version
 *
 **/
SPROX_API_FUNC2(const TCHAR *, GetErrorMessage) (SWORD status)
{
  static TCHAR msg[256];
  switch (status)
  {
    case MI_OK:
      return _T("function was executed without failure");
    case MI_NOTAGERR:
      return _T("no card in the operating area of the antenna, or not authenticated");
    case MI_CRCERR:
      return _T("card's CRC was not correct");
    case MI_EMPTY:
      return _T("value overflow during increment or underflow during decrement");
    case MI_AUTHERR:
      return _T("no card authentication possible due to invalid CRYPTO 1 keys");
    case MI_PARITYERR:
      return _T("card's parity was not correct");
    case MI_CODEERR:
      return _T("NACK received, card detected CRC or parity error");
    case MI_SERNRERR:
      return _T("check byte of card's serial number (got from anticollision) is not correct");
    case MI_KEYERR:
      return _T("authentication key error");
    case MI_NOTAUTHERR:
      return _T("card's sector is not authenticated");
    case MI_BITCOUNTERR:
      return _T("wrong number of bits received from the card");
    case MI_BYTECOUNTERR:
      return _T("wrong number of bytes received from the card");
    case MI_IDLE:
      return _T("idle command active");
    case MI_TRANSERR:
      return _T("transfert operation denied");
    case MI_WRITEERR:
      return _T("write operation denied");
    case MI_INCRERR:
      return _T("increment operation denied");
    case MI_DECRERR:
      return _T("decrement operation denied");
    case MI_READERR:
      return _T("read operation denied");
    case MI_OVFLERR:
      return _T("FIFO overflow error");
    case MI_POLLING:
      return _T("message after initialising the polling mode");
    case MI_FRAMINGERR:
      return _T("framing error of the RF interface, start bit not valid");
    case MI_ACCESSERR:
      return _T("timeout accessing the reader module");
    case MI_UNKNOWN_COMMAND:
      return _T("unknown command");
    case MI_COLLERR:
      return _T("collision during RF data transfer");
    case MI_RESETERR:
      return _T("error during reset/initialisation of the RC5xx");
    case MI_INTERFACEERR:
      return _T("MF RC5xx is not responding correctly");
    case MI_ACCESSTIMEOUT:
      return _T("MF RC5xx is not responding in time");
    case MI_NOBITWISEANTICOLL:
      return _T("card does not support bitwise anticollision");

    case MI_QUIT:
      return _T("MI_QUIT !");
    case MI_CODINGERR:
      return _T("wrong coding of ACK or NACK code");

    case MI_SENDBYTENR:
      return _T("T=CL wrong send byte number");
    case MI_CASCLEVEX:
      return _T("wrong casc level !");
    case MI_SENDBUF_OVERFLOW:
      return _T("T=CL send buffer overflow");
    case MI_BAUDRATE_NOT_SUPPORTED:
      return _T("wrong baud rate selected");
    case MI_SAME_BAUDRATE_REQUIRED:
      return _T("receiving and sending with the same baud rate");
    case MI_WRONG_PARAMETER_VALUE:
      return _T("wrong parameter was passed to the function");

    case MI_BREAK:
      return _T("break !");
    case MI_NY_IMPLEMENTED:
      return _T("this function is not available or not implemented");
    case MI_NO_MFRC:
      return _T("the hardware version is not 2.x");
    case MI_MFRC_NOTAUTH:
      return _T("MF RC5xx is not authenticated (and therefore denies this function)");
    case MI_WRONG_DES_MODE:
      return _T("the selected DES mode does not match the MF RC5xx configuration");
    case MI_HOST_AUTH_FAILED:
      return _T("the MF RC5xx rejects the host during host authentication");

    case MI_WRONG_LOAD_MODE:
      return _T("the mode for master key and access condition loading does not match the MF RC5xx configuration");
    case MI_WRONG_DESKEY:
      return _T("the least significant bits of the DES key bytes are not zero");
    case MI_MKLOAD_FAILED:
      return _T("master keys or access conditions were rejected by the MF RC5xx");
    case MI_FIFOERR:
      return _T("error during FIFO handling");
    case MI_WRONG_ADDR:
      return _T("the specified address is out of range");
    case MI_DESKEYLOAD_FAILED:
      return _T("DES keys were rejected by the MF RC5xx");
    case MI_RECBUF_OVERFLOW:
      return _T("Receive buffer overflow");

    case MI_WRONG_SEL_CNT:
      return _T("this select counter is not valid");

    case MI_WRONG_TEST_MODE:
      return _T("this test mode is not valid");
    case MI_TEST_FAILED:
      return _T("the self-test routine failed");
    case MI_TOC_ERROR:
      return _T("the timeout counter did not behave as expected");
    case MI_COMM_ABORT:
      return _T("communication aborted");
    case MI_INVALID_BASE:
      return _T("the base address is not in the range of 0x0200 to 0x03F0");
    case MI_MFRC_RESET:
      return _T("MF RC5xx is busy with start up or it is in reset");
    case MI_WRONG_VALUE:
      return _T("wrong value passed");
    case MI_VALERR:
      return _T("value format error / invalid data for counter");
    case MI_WRONG_LENGTH:
      return _T("wrong length for command parameters");

    case MI_FUNCTION_NOT_AVAILABLE:
      return _T("SProx : the function is not available");
    case MI_TIME_EXTENSION:
      return _T("SProx : reader goes on working (not an error, update the library !)");

    case MI_SER_LENGTH_ERR:
      return _T("SProx : length of received message is not correct");
    case MI_SER_CHECKSUM_ERR:
      return _T("SProx : CRC of received message is not correct");
    case MI_SER_PROTO_ERR:
      return _T("SProx : synchronization lost in received message");
    case MI_SER_PROTO_NAK:
      return _T("SProx : NACK received");
    case MI_SER_ACCESS_ERR:
      return _T("SProx : access error");
    case MI_SER_TIMEOUT_ERR:
      return _T("SProx : timeout");
    case MI_SER_NORESP_ERR:
      return _T("SProx : no response");

    case MI_LIB_CALL_ERROR:
      return _T("SProx : invalid call to library function");
    case MI_OUT_OF_MEMORY_ERROR:
      return _T("SProx : out of memory in library function");
    case MI_FILE_ACCESS_ERROR:
      return _T("SProx : file error in library function");

    case TCL_NOTAGERR:
      return _T("no T=CL card in the operating area of the antenna, or no card selected");
    case TCL_CRCERR:
      return _T("T=CL : card's CRC was not correct");
    case TCL_PRITYERR:
      return _T("T=CL : card's parity was not correct");
    case TCL_OTHERERR:
      return _T("T=CL : other card error");
    case TCL_SERNRERR:
      return _T("T=CL : check byte of card's serial number is not correct");
    case TCL_BITCOUNTERR:
      return _T("T=CL : wrong number of bits received from the card");
    case TCL_POLLING:
      return _T("T=CL : message after initialising the polling mode");
    case TCL_RF_CHANNEL:
      return _T("T=CL : error on RF channel");
    case TCL_MULTACT_DISABLED:
      return _T("T=CL : multi-activation is disabled");
    case TCL_MULTACT_ENABLED:
      return _T("T=CL : multi-activation is enabled");
    case TCL_CID_NOT_ACTIVE:
      return _T("T=CL : no active card with this CID");
    case TCL_BITANTICOLL:
      return _T("T=CL : bitwise anticollision detected");
    case TCL_UIDLEN:
      return _T("T=CL : wrong UID len, or not supported");
    case TCL_CIDINVALID:
      return _T("T=CL : invalid CID, or not supported");
    case TCL_ATSLEN:
      return _T("T=CL : wrong ATS length, or unsupported value");
    case TCL_NO_ATS_AVAILABLE:
      return _T("T=CL : no ATS available (card is not 14443-4 ?)");
    case TCL_ATS_ERROR:
      return _T("T=CL : error while retrieving card ATS, or invalid ATS");
    case TCL_FATAL_PROTOCOL:
      return _T("T=CL : fatal protocol error");
    case TCL_RECBUF_OVERFLOW:
      return _T("T=CL : receive buffer overflow");
    case TCL_SENDBYTENR:
      return _T("T=CL : receive buffer overflow");
    case TCL_TRANSMERR_HALTED:
      return _T("T=CL : transmit halted");
    case TCL_TRANSMERR_NOTAG:
      return _T("T=CL : tag removed while transfering");
    case TCL_BAUDRATE_NOT_SUPPORTED_PICC:
      return _T("T=CL : baudrate not supported by the card");
    case TCL_CID_NOT_SUPPORTED:
      return _T("T=CL : the card doesn't support the CID feature");
    case TCL_NAD_NOT_SUPPORTED:
      return _T("T=CL : the card doesn't support the NAD feature");
    case TCL_PROTOCOL_NOT_SUPPORTED:
      return _T("T=CL : the card doesn't support the T=CL protocol");
    case TCL_PPS_FORMAT:
      return _T("T=CL : wrong PPS frame format");
    case TCL_PPS_ERROR:
      return _T("T=CL : PPS response error");
    case TCL_NADINVALID:
      return _T("T=CL : invalid NAD, or not supported");
    case TCL_OTHER_ERR:
      return _T("T=CL : other protocol error");
    case TCL_BAUDRATE_NOT_SUPPORTED_PCD:
      return _T("T=CL : baudrate not supported by the reader");
    case TCL_CID_ACTIVE:
      return _T("T=CL : this CID is active");
      
    case VIC_CUSTOMERR:
      return _T("15693 : custom command error codes not implemented");
    case VIC_CMD_OPTION_NOT_SUPPORTED:
      return _T("15693 : option not supported");
    case VIC_ERR:
      return _T("15693 : error with no information given");
    case VIC_BLOCK_NOT_EXIST:
      return _T("15693 : the specified block is not available (doesnt exist)");
    case VIC_BLOCK_ALREADY_LOCKED:
      return _T("15693 : the specified block is already locked and thus cannot be locked again");
    case VIC_BLOCK_LOCKED:
      return _T("15693 : the specified block is locked and its content cannot be changed");
    case VIC_BLOCK_PROGRAMERR:
      return _T("15693 : the specified block was not successfully programmed");
    case VIC_BLOCK_LOCKERR:
      return _T("15693 : the specified block was not successfully locked");
      
    case I1_TIMESLOTOCCUPIED:
      return _T("ICODE1 : timeslot is occupied with another card");
    case I1_NO_HALT:
      return _T("ICODE1 : card was not successfully halted");

  }

  _stprintf(msg, _T("Unknown error status %d"), status);
  return msg;
}

SPROX_API_FUNC2(const wchar_t *, GetErrorMessageW) (SWORD status)
{
#ifdef UNICODE
  return SPROX_GetErrorMessage(status);
#else
  static wchar_t buffer[256+1];
  const char *p;
  int i;
  p = SPROX_GetErrorMessage(status);
  for (i=0; i<256; i++)
  {
    buffer[i] = p[i];
    if (buffer[i] == '\0') break;
  }
  buffer[256] = '\0';
  return buffer;  
#endif  
}

SPROX_API_FUNC2(const char *, GetErrorMessageA) (SWORD status)
{
#ifndef UNICODE
  return SPROX_GetErrorMessage(status);
#else
  static char buffer[256+1];
  const wchar_t *p;
  int i;
  p = SPROX_GetErrorMessage(status);
  for (i=0; i<256; i++)
  {
    buffer[i] = (char) p[i];
    if (buffer[i] == '\0') break;
  }
  buffer[256] = '\0';
  return buffer;  
#endif  
}

/**f* SpringProxAPI/SPROX_ArrayToString
 *
 * NAME
 *   SPROX_ArrayToString
 *
 * DESCRIPTION
 *   Translate an array of bytes into a string of hexadecimal digits
 *
 * INPUTS
 *   TCHAR *string      : buffer to hold the hexadecimal string
 *   const BYTE *buffer : byte array
 *   WORD size          : size of byte array
 *
 * RETURNS
 *   MI_OK              : success
 *
 * NOTES
 *   You must ensure that size of string buffer is enough to store
 *   the whole hexadecimal string (i.e. at least 2 * size characters)
 *
 * VARIANTS
 *   SPROX_ArrayToStringA : ASCII version
 *   SPROX_ArrayToStringW : UNICODE version
 *
 **/
SPROX_API_FUNC(ArrayToString) (TCHAR *string, const BYTE *buffer, WORD size)
{
#ifdef UNICODE
  return SPROX_ArrayToStringW(string, buffer, size);
#else
  return SPROX_ArrayToStringA(string, buffer, size);
#endif
}

#ifdef WIN32
SPROX_API_FUNC(ArrayToStringW) (wchar_t *string, const BYTE *buffer, WORD size)
{
  WORD    i;
  wchar_t tmp[3];
  if (string == NULL) return MI_LIB_CALL_ERROR;
  if (buffer == NULL) return MI_LIB_CALL_ERROR;
  string[0] = '\0';
  for (i = 0; i < size; i++)
  {
    swprintf(tmp, L"%02X", buffer[i]);
    wcscat(string, tmp);
  }
  return MI_OK;
}
#endif
SPROX_API_FUNC(ArrayToStringA) (char *string, const BYTE *buffer, WORD size)
{
  WORD i;
  char tmp[3];
  if (string == NULL) return MI_LIB_CALL_ERROR;
  if (buffer == NULL) return MI_LIB_CALL_ERROR;
  string[0] = '\0';
  for (i = 0; i < size; i++)
  {
    sprintf(tmp, "%02X", buffer[i]);
    strcat(string, tmp);
  }
  return MI_OK;
}

/**f* SpringProxAPI/SPROX_StringToArray
 *
 * NAME
 *   SPROX_StringToArray
 *
 * DESCRIPTION
 *   Translate a string of hexadecimal digits into an array of bytes
 *
 * INPUTS
 *   BYTE *buffer        : byte array to receive the result
 *   const TCHAR *string : hexadecimal string
 *   WORD size           : size of the byte array
 *
 * RETURNS
 *   MI_OK              : success
 *
 * NOTES
 *   The length of the string must be equals to 2 * size
 *
 * VARIANTS
 *   SPROX_StringToArrayA : ASCII version
 *   SPROX_StringToArrayW : UNICODE version
 *
 **/
static BYTE h2q(char h)
{
  if ((h >= '0') && (h <= '9')) return h - '0';
  if ((h >= 'A') && (h <= 'F')) return h - 'A' + 10;
  if ((h >= 'a') && (h <= 'f')) return h - 'a' + 10;
  return 0xFF;
}

SPROX_API_FUNC(StringToArray) (BYTE *buffer, const TCHAR *string, WORD size)
{
#ifdef UNICODE
  return SPROX_StringToArrayW(buffer, string, size);
#else
  return SPROX_StringToArrayA(buffer, string, size);
#endif
}

#ifdef WIN32
SPROX_API_FUNC(StringToArrayW) (BYTE *buffer, const wchar_t *string, WORD size)
{
  WORD i;
  BYTE q;

  if (string == NULL) return MI_LIB_CALL_ERROR;
  if (buffer == NULL) return MI_LIB_CALL_ERROR;
  if (wcslen(string) < (size_t) (2 * size)) return MI_LIB_CALL_ERROR;

  for (i = 0; i < size; i++)
  {
    q = h2q((BYTE) string[2 * i]);
    if (q > 0x0F) return MI_LIB_CALL_ERROR;
    buffer[i] = q;
    buffer[i] <<= 4;
    q = h2q((BYTE) string[2 * i + 1]);
    if (q > 0x0F) return MI_LIB_CALL_ERROR;
    buffer[i] += q;
  }

  return MI_OK;
}
#endif

SPROX_API_FUNC(StringToArrayA) (BYTE *buffer, const char *string, WORD size)
{
  WORD i;
  BYTE q;

  if (string == NULL) return MI_LIB_CALL_ERROR;
  if (buffer == NULL) return MI_LIB_CALL_ERROR;
  if (strlen(string) < (size_t) (2 * size)) return MI_LIB_CALL_ERROR;

  for (i = 0; i < size; i++)
  {
    q = h2q((BYTE) string[2 * i]);
    if (q > 0x0F) return MI_LIB_CALL_ERROR;
    buffer[i] = q;
    buffer[i] <<= 4;
    q = h2q((BYTE) string[2 * i + 1]);
    if (q > 0x0F) return MI_LIB_CALL_ERROR;
    buffer[i] += q;
  }

  return MI_OK;
}

/**f* SpringProxAPI/SPROX_StrLen
 *
 * NAME
 *   SPROX_StrLen
 *
 * DESCRIPTION
 *   Determines the length of a string
 *
 * INPUTS
 *   const TCHAR *string : input string
 *
 * RETURNS
 *   WORD : actual length (= number of characters) of the string
 *
 * VARIANTS
 *   SPROX_StrLenA : ASCII version
 *   SPROX_StrLenW : UNICODE version
 *
 **/
SPROX_API_FUNC2(WORD, StrLen) (const TCHAR *string)
{
  if (string == NULL) return 0;
  return (WORD) _tcsclen(string);
}

#ifdef WIN32
SPROX_API_FUNC2(WORD, StrLenW) (const wchar_t *string)
{
  if (string == NULL)  return 0;
  return (WORD) wcslen(string);
}
#endif
SPROX_API_FUNC2(WORD, StrLenA) (const char *string)
{
  if (string == NULL)  return 0;
  return (WORD) strlen(string);
}

/**f* SpringProxAPI/SPROX_Malloc
 *
 * NAME
 *   SPROX_Malloc
 *
 * DESCRIPTION
 *   Dynamic allocation of a buffer
 *
 * INPUTS
 *   BYTE **buffer      : pointer to receive the address of the allocated buffer
 *   WORD size          : size of the buffer
 *
 * RETURNS
 *   MI_OK              : success
 *
 * NOTES
 *   The buffer must be freed using SPROX_Free
 *
 **/
SPROX_API_FUNC(Malloc) (BYTE ** buffer, WORD size)
{
  if (buffer == NULL)
    return MI_LIB_CALL_ERROR;
  *buffer = (BYTE *) malloc(size);
  if (*buffer == NULL)
    return MI_OUT_OF_MEMORY_ERROR;
  return MI_OK;
}

/**f* SpringProxAPI/SPROX_Free
 *
 * NAME
 *   SPROX_Free
 *
 * DESCRIPTION
 *   Desallocation of a dynamic buffer
 *
 * INPUTS
 *   BYTE *buffer       : address of the allocated buffer
 *
 * RETURNS
 *   MI_OK              : success
 *
 * SEE ALSO
 *   SPROX_Malloc
 *
 **/
SPROX_API_FUNC(Free) (BYTE *buffer)
{
  if (buffer == NULL)
    return MI_LIB_CALL_ERROR;
  free(buffer);
  return MI_OK;
}

#endif

#endif
