#include "sprox_desfire_i.h"

void dump_buffer(const BYTE * buffer, WORD len)
{
  WORD    i;
  for (i = 0; i < len; i++)
  {
    printf("%02X", buffer[i]);
    if (i % 8 == 7)
      printf(" ");
    if (i % 64 == 63)
      printf("\n");
  }
}

/*
 **************************************************************************************************
 *
 *  Exported functions
 *
 *************************************************************************************************
 */
SPROX_DESFIRE_FUNC_EX(Desfire_SelectCid) (SPROX_PARAM  BYTE cid)
{
  SPROX_DESFIRE_GET_CTX();
  ctx->cid = cid;
  return 0;
}

/*
 **************************************************************************************************
 *
 *  Command : "low" level command<->response transfert
 *
 *************************************************************************************************
 */

SPROX_DESFIRE_FUNC_EX(Desfire_Command) (SPROX_PARAM  BOOL fast, WORD accept_len, BYTE accept_status)
{
  SWORD   status;
  BYTE    card_status;
  WORD    len = DESFIRE_BUFFER_SIZE;
  SPROX_DESFIRE_GET_CTX();

//  D(printf("S: "));
//  D(dump_buffer(ctx->buffer, ctx->buflen);
//  D(printf("\n"));

  /* Communicate the prepared buffer to the PICC and get its response. */
  if (fast)
  {
    status = SPROX_API_CALL(TclA_ExchangeDF) (SPROX_PARAM_P  ctx->cid, ctx->buffer, ctx->buflen, ctx->buffer, &len);
  } else
  {
    status = SPROX_API_CALL(TclA_Exchange) (SPROX_PARAM_P  5, ctx->cid, 0xFF, ctx->buffer, ctx->buflen, ctx->buffer, &len);
  }
  if (status != MI_OK)
  {
    /* Error: could not exchange info blocks with the PICC. */
    return status;
  }

  /* Remember received size */
  ctx->buflen = len;

//  D(printf("R: "));
//  D(dump_buffer(ctx->buffer, ctx->buflen);
//  D(printf("\n"));

  /* Check whether the PICC's response consists of at least one byte.
     If the response is empty, we can not even determine the PICC's status.
     Therefore an empty response is always a length error. */
  if (len < 1)
  {
    /* Error: block with inappropriate number of bytes received from the PICC. */
    return DFCARD_WRONG_LENGTH;
  }

  /*  Check the status byte received from the PICC.
     Depending on which of the flags (sgbACCEPT_OPERATION_OK and sgbACCEPT_ADDITIONAL_FRAME)
     are set in the bExpectedStatusFlags parameter, we accept either the status byte 
     sgbOPERATION_OK or the status byte sgbDF_ADDITIONAL_FRAME or both of them. */
  card_status = ctx->buffer[INF + 0];

  if (!(((card_status == DF_OPERATION_OK) && (accept_status & ACCEPT_OPERATION_OK)) ||
        ((card_status == DF_ADDITIONAL_FRAME) && (accept_status & ACCEPT_ADDITIONAL_FRAME))))
  {
    /* Error: error or unexpected status received from the PICC. */
    return (DFCARD_ERROR - card_status);
  }

  /* Finally check the length of the response. It must be > 0 (already checked) and
     <= sgbMAX_INFO_FRAME_SIZE in all cases.
     If bExpectedResponseLength is != 0, we expect a response with exactly that length.
     If bExpectedResponseLength == 0, we allow any response length > 0 and 
     <= sgbMAX_INFO_FRAME_SIZE. */
  if (((!fast) && (len > MAX_INFO_FRAME_SIZE)) || ((accept_len != 0) && (accept_len != len)))
  {
    /* Error: block with inappropriate number of bytes received from the PICC. */
    return DFCARD_WRONG_LENGTH;
  }

  /* Return the status code from the PICC without any prefix code. */
  return status;
}



/*
 **************************************************************************************************
 *
 *  ReadDataEx / WriteDataEx / ModifyValue internal functions
 *
 *************************************************************************************************
 */

/*
 * ReadDataEx
 * ----------
 */
SPROX_DESFIRE_FUNC_EX(Desfire_ReadDataEx) (SPROX_PARAM  BYTE read_command, BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD item_count, DWORD item_size, BYTE data[], DWORD *done_size)
{
  SWORD   status;
  DWORD   recv_len = 0;
  DWORD   padd_pos = 0;
  DWORD   byte_count;
  DWORD   temp;
#ifndef UNDER_CE
  static BYTE buffer[DESFIRE_BUFFER_SIZE + 16];
#else
  BYTE    buffer[DESFIRE_BUFFER_SIZE + 16];
#endif
  SPROX_DESFIRE_GET_CTX();
  
  /* We have to calculate the number of bytes as this function works at byte granularity.
     Cast the DWORD to double in the multiplication to detect an overflow in the result.
     If an overflow occurs, continue with the maximum number provided by a DWORD. The PICC
     will refuse the read request in this case with a BOUNDARY_ERROR. */
  if (item_count * item_size > 4096)
    byte_count = 0xFFFFFFFF;
  else
    byte_count = item_count * item_size;

  ctx->buffer[INF + 0] = read_command;
  /* we do not have to take the item_size into account for the 
     generation of the command header as the PICC resolves the meaning of this
     parameter due to the specific command code */
  ctx->buffer[INF + 1] = file_id;
  temp = from_offset;
  ctx->buffer[INF + 2] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 3] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (temp & 0x000000FF);
  temp = item_count;
  ctx->buffer[INF + 5] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (temp & 0x000000FF);
  ctx->buflen = 8;

  for (;;)
  {
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  TRUE, 0, DF_ADDITIONAL_FRAME | ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
    {
      return status;
    }
    memcpy(&buffer[recv_len], &ctx->buffer[INF + 1], ctx->buflen - 1);
    recv_len += (ctx->buflen - 1);
    if (ctx->buffer[INF + 0] != DF_ADDITIONAL_FRAME)
      break;
    ctx->buflen = 1;
  }

  /* correct termination of this loop only is possible if we received
     a frame with the status code sgbOPERATION_OK
     this means that this was the last frame sent by the PICC
     the additional frame status code which was prepared at the end of the loop 
     is not sent but discarded */

  /* decide upon the communications mode which cryptographic operation
     is to be applied on the received data */
  if ((comm_mode == DF_COMM_MODE_PLAIN) || (comm_mode == DF_COMM_MODE_PLAIN2))
  {
    if (item_count == 0)
    {
      /* if no length parameter has been passed, 
         eNBytes is set with the number of received bytes
         this is valid for both, record files and data files */
      byte_count = recv_len;
    } else
    {
      /* if a length parameter has been passed,
         we check, whether we got exactly that much data. */
      if (recv_len != byte_count)
      {
        /* length error */
        return DFCARD_WRONG_LENGTH;
      }
    }
  } else if (comm_mode == DF_COMM_MODE_MACED)
  {
    BYTE    mac32[4];

    SPROX_API_CALL(Desfire_CipherMACb) (SPROX_PARAM_P  buffer, (WORD) (recv_len - 4), ctx->session_key, mac32);
    if (memcmp(mac32, &buffer[recv_len - 4], 4))
    {
      /* abortion due to integrity error -> wrong MAC */
      return DFCARD_WRONG_MAC;
    }
    if (item_count == 0)
    {
      /* if no length parameter has been passed 
         eNBytes is set with the number of received bytes
         this is valid for both, record files and data files */
      byte_count = recv_len - 4;
    }
  } else if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* check whether the received data count is multiple of 8 */
    if (recv_len % 8)
    {
      return DF_LENGTH_ERROR;
    }

    /* at first we have to decipher the buffer */
    SPROX_API_CALL(Desfire_CipherRecv) (SPROX_PARAM_P  buffer, (WORD) (recv_len / 8), ctx->session_key);

    /* now there are two different paths caused by the two different padding methods */

    /* if the data length was specified in the parameter section of the command
       the buffer content is padded with all zeroes or possibly none,
       after the CRC to reach multiples of 8 bytes, suitable for the DesOperation 
       in the other case, where no data length was specified, the buffer content is padded
       at least with one byte coded to 0x80 which is followed by all zeroes or possibly 
       none, after the CRC to reach multiples of 8 bytes, suitable for the DesOperation */
    if (item_count != 0)
    {
      /* set the starting position for checking the zero padding bytes (after CRC) */
      padd_pos = byte_count + 2;
    } else
    {
      /* no length was specified, therefore the second padding rule applies
         we have to find the character 0x80 at the end of the received data
         and obtain the number of data bytes by subtracting 2 bytes from the position
         where 0x80 was found ( position + 1 = length ) */
      for (byte_count = (recv_len - 1); byte_count > 0; byte_count--)
      {
        if (buffer[byte_count] == 0x80)
        {
          byte_count = byte_count - 2;
          break;
        } else if (buffer[byte_count] != 0x00 || byte_count == 1)
        {
          return DFCARD_WRONG_PADDING;
        }
      }
      padd_pos = byte_count + 3;
    }

    /* this is the CRC check for both cases (specified and unspecified length) */
    if (SPROX_ComputeIso14443ACrc(NULL, buffer, (WORD) (byte_count + 2)))
    {
      /* abortion due to integrity error -> wrong CRC */
      return DFCARD_WRONG_CRC;
    }

    /*  Finally we do a check on the rest of the padding bytes ( if there are any )
       which have to be all zero for both cases */

    /* note: in case of the unspecified length parameter, there must be at least
       the 0x80 byte ( has been checked already! ) from there on
       both cases are again the same */
    if (recv_len > byte_count + 2)
    {
      for (; padd_pos < recv_len; padd_pos++)
      {
        if (buffer[padd_pos] != 0x00)
        {
          return DFCARD_WRONG_PADDING;
        }
      }
    }
  }

  /* return the number of bytes if a pointer was passed */
  if (done_size != NULL)
    *done_size = byte_count;

  /* copy data */
  if (data != NULL)
    memcpy(data, buffer, byte_count);

  return DF_OPERATION_OK;
}

/*
 * WriteDataEx
 * -----------
 */
SPROX_DESFIRE_FUNC_EX(Desfire_WriteDataEx) (SPROX_PARAM  BYTE write_command, BYTE file_id, BYTE comm_mode, DWORD from_offset, DWORD size, const BYTE data[])
{
  DWORD   done_size;
  WORD    next_size;
  DWORD   temp;
  SWORD   status;
#ifndef UNDER_CE
  static BYTE buffer[4096 + 16];
#else
  BYTE    buffer[4096 + 16];
#endif
  SPROX_DESFIRE_GET_CTX();

  memset(buffer, 0xCC, sizeof(buffer)); // TODO
  memcpy(buffer, data, size);

  done_size = 0;

  ctx->buffer[INF + 0] = write_command;
  ctx->buffer[INF + 1] = file_id;
  temp = from_offset;
  ctx->buffer[INF + 2] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 3] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (temp & 0x000000FF);
  temp = size;
  ctx->buffer[INF + 5] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 6] = (BYTE) (temp & 0x000000FF);
  temp >>= 8;
  ctx->buffer[INF + 7] = (BYTE) (temp & 0x000000FF);
  ctx->buflen = 8;

  /* decide upon the communications mode which cryptographic */
  /* operation is to be applied on the data                  */

  if (comm_mode == DF_COMM_MODE_MACED)
  {
    /* append the 4 bytes MAC */
    SPROX_API_CALL(Desfire_CipherMACb) (SPROX_PARAM_P  buffer, (WORD) size, ctx->session_key, &buffer[size]);
    size += 4;
  } else if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* at first we have to append the 2 bytes CRC */
    SPROX_ComputeIso14443ACrc(&buffer[size], buffer, (WORD) size);
    size += 2;

    while (size % 8)
      buffer[size++] = 0;

    /* finally we do the cipher operation on the modified data */
    SPROX_API_CALL(Desfire_CipherSend) (SPROX_PARAM_P  buffer, (WORD) (size / 8), ctx->session_key);
  }

  /* if comm_mode is neither MACed nor ciphered we leave the data as it is */
  /* this means a plain communication                                      */

  /* Now our data is ready for being sent to the PICC */

  do
  {
    /* determine the limiting factor for the number of bytes to be appended to     */
    /* the ctx->buffer                                                             */
    /* this is either the maximum frame size or the number of bytes which are left */
    if ((size - done_size) > (DWORD) (DF_MAX_INFO_FRAME_SIZE - ctx->buflen))
    {
      /* transfer the calculated number of bytes  */
      next_size = (WORD) (DF_MAX_INFO_FRAME_SIZE - ctx->buflen);
    } else
    {
      /* transfer all the bytes left  */
      next_size = (WORD) (size - done_size);
    }

    /* Now put these bytes into the ctx->buffer */
    memcpy(&ctx->buffer[INF + ctx->buflen], &buffer[done_size], next_size);
    ctx->buflen += (WORD) next_size;

    /* only one byte response is allowed                                */
    /* allowed status codes are DF_ADDITIONAL_FRAME and DF_OPERATION_OK */
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 1, DF_ADDITIONAL_FRAME | ACCEPT_OPERATION_OK);
    if (status != DF_OPERATION_OK)
    {
      return status;
    }

    done_size += (DWORD) next_size;

    /* put the ADDITONAL_FRAME status code for the next frame into the buffer */
    ctx->buffer[INF + 0] = DF_ADDITIONAL_FRAME;
    ctx->buflen = 1;
  } while (done_size < size);

  /* because ( eNumOfBytesExtracted = eNBytesToWrite ) is the only way for     */
  /* leaving the loop correctly an interrupted write operation is detected via */
  /* a status code different from DF_OPERATION_OK                              */
  return status;
}

/*
 * ModifyValue
 * -----------
 */
SPROX_DESFIRE_FUNC_EX(Desfire_ModifyValue) (SPROX_PARAM  BYTE modify_command, BYTE file_id, BYTE comm_mode, SDWORD amount)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code, the file ID argument and the */
  /* modification amount argument.                                                   */
  ctx->buffer[INF + 0] = modify_command;
  ctx->buffer[INF + 1] = file_id;
  ctx->buffer[INF + 2] = (BYTE) (amount & 0x000000FF);
  amount >>= 8;
  ctx->buffer[INF + 3] = (BYTE) (amount & 0x000000FF);
  amount >>= 8;
  ctx->buffer[INF + 4] = (BYTE) (amount & 0x000000FF);
  amount >>= 8;
  ctx->buffer[INF + 5] = (BYTE) (amount & 0x000000FF);
  ctx->buflen = 6;

  /* Process the amount argument according to the selected communication mode. */
  if (comm_mode == DF_COMM_MODE_ENCIPHERED)
  {
    /* Append a CRC to the amount argument. */
    SPROX_ComputeIso14443ACrc(&ctx->buffer[INF + 6], &ctx->buffer[INF + 2], 4);

    /* 'Encrypt' the amount argument bytes and the CRC bytes padding         */
    /* the DES block with '0' bytes.                                         */
    /* Note: since mifare DESFire always runs in DES ENcryption mode this is */
    /* actually a DES DEcryption operation.                                  */
    ctx->buffer[INF + 8] = 0;
    ctx->buffer[INF + 9] = 0;

    SPROX_API_CALL(Desfire_CipherSend) (SPROX_PARAM_P  &ctx->buffer[INF + 2], 1, ctx->session_key);
    ctx->buflen += 4;
  } else if (comm_mode == DF_COMM_MODE_MACED)
  {
    /* Append a MAC to the amount argument */
    SPROX_API_CALL(Desfire_CipherMACb) (SPROX_PARAM_P  &ctx->buffer[INF + 2], 4, ctx->session_key, &ctx->buffer[INF + 6]);
    ctx->buflen += 4;
  }

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  FALSE, 0, ACCEPT_OPERATION_OK);
}
