#include "cardappz.h"


int JaycosSelectCorpid(CARDAPPZ_CTX_ST * ctx)
{
  int rc;
  unsigned char key[16];

  /* Select the application directory */
  rc = JaycosSelectDF(ctx, CORPID_JAYCOS_AID);
  if (rc)
    return rc;

  /* Compute the diversified key 1 */
  rc = ComputeDivKey(ctx, key, CORPID_MASTER_KEY1);
  if (rc)
    return rc;

  /* Get authenticated with key 1 */
  rc = JaycosMutualAuthenticate(ctx, CORPID_JAYCOS_KEY1_ID, key);
  if (rc)
    return rc;

  return CARDAPPZ_OK;
}

int JaycosReadCorpidIdentity(CARDAPPZ_CTX_ST * ctx,
                             CORPID_IDENTITY_ST * identity)
{
  return JaycosReadBinary(ctx, CORPID_JAYCOS_FID_IDENTITY, 0,
                          sizeof(CORPID_IDENTITY_ST),
                          (unsigned char *) identity);
}

int JaycosReadCorpidFingerprints(CARDAPPZ_CTX_ST * ctx,
                                 unsigned char
                                 fingerprints[CORPID_FINGERPRINTS_MAX_SZ])
{
  return JaycosReadBinary(ctx, CORPID_DESFIRE_FID_FINGERPRINTS, 0,
                          CORPID_FINGERPRINTS_MAX_SZ, fingerprints);
}

int JaycosReadCorpidPicture(CARDAPPZ_CTX_ST * ctx,
                            unsigned char picture[CORPID_PICTURE_MAX_SZ])
{
  return JaycosReadBinary(ctx, CORPID_JAYCOS_FID_PICTURE, 0,
                          CORPID_PICTURE_MAX_SZ, picture);
}

int JaycosReadCorpidAccessControl(CARDAPPZ_CTX_ST * ctx,
                                  unsigned char accesscontrol[48])
{
  return JaycosReadBinary(ctx, CORPID_JAYCOS_FID_ACCESSCONTROL, 0, 48,
                          accesscontrol);
}

int JaycosReadCorpidOptionsSign(CARDAPPZ_CTX_ST * ctx,
                                CORPID_OPTIONSSIGN_ST * optionssing)
{
  return JaycosReadBinary(ctx, CORPID_JAYCOS_FID_OPTIONSSIGN, 0,
                          sizeof(CORPID_OPTIONSSIGN_ST),
                          (unsigned char *) optionssing);
}

int JaycosWriteCorpid(CARDAPPZ_CTX_ST * ctx, CORPID_IDENTITY_ST * identity,
                      unsigned char fingerprints[CORPID_FINGERPRINTS_MAX_SZ],
                      unsigned char picture[CORPID_PICTURE_MAX_SZ],
                      unsigned char accesscontrol[48],
                      CORPID_OPTIONSSIGN_ST * optionssing)
{
  int rc;
  unsigned char key[16];

  /* Select the application directory */
  rc = JaycosSelectDF(ctx, CORPID_JAYCOS_AID);
  if (rc)
    return rc;

  /* Compute the diversified key 0 */
  rc = ComputeDivKey(ctx, key, CORPID_MASTER_KEY0);
  if (rc)
    return rc;

  /* Get authenticated with key 0 */
  rc = JaycosMutualAuthenticate(ctx, CORPID_JAYCOS_KEY0_ID, key);
  if (rc)
    return rc;

  if (identity != NULL)
  {
    rc =
      JaycosUpdateBinary(ctx, CORPID_JAYCOS_FID_IDENTITY, 0,
                         sizeof(CORPID_IDENTITY_ST),
                         (unsigned char *) identity);
    if (rc)
      return rc;
  }

  if (fingerprints != NULL)
  {
    rc =
      JaycosUpdateBinary(ctx, CORPID_JAYCOS_FID_FINGERPRINTS, 0,
                         CORPID_FINGERPRINTS_MAX_SZ, fingerprints);
    if (rc)
      return rc;
  }

  if ((picture != NULL) && (optionssing != NULL))
  {
    /* Size of picture must be retrieved in options */
    unsigned short size = CtoHs(optionssing->PictureSize);

    rc = JaycosUpdateBinary(ctx, CORPID_JAYCOS_FID_PICTURE, 0, size, picture);
    if (rc)
      return rc;
  }

  if (accesscontrol != NULL)
  {
    rc =
      JaycosUpdateBinary(ctx, CORPID_JAYCOS_FID_ACCESSCONTROL, 0, 48,
                         accesscontrol);
    if (rc)
      return rc;
  }

  if (optionssing != NULL)
  {
    rc =
      JaycosUpdateBinary(ctx, CORPID_JAYCOS_FID_OPTIONSSIGN, 0,
                         sizeof(CORPID_OPTIONSSIGN_ST),
                         (unsigned char *) optionssing);
    if (rc)
      return rc;
  }

  return CARDAPPZ_OK;
}

/* Formatting */

/* ---------- */
int JaycosFormatCorpid(CARDAPPZ_CTX_ST * ctx)
{
  int rc;
  unsigned char key[16];

  /* Select application 0 */
  rc = JaycosSelectMF(ctx);
  if (rc)
    return rc;

  /* Create the identity application */
  rc = JaycosCreateDF(ctx, CORPID_JAYCOS_AID);
  if (rc)
    return rc;

  /* Select the application */
  rc = JaycosSelectDF(ctx, CORPID_JAYCOS_AID);
  if (rc)
    return rc;

  /* Formatting is performed with master key */
  rc = JaycosExternalAuthenticate(ctx, 0, JAYCOS_MASTER_KEY);
  if (rc)
    return rc;

  /* Create user data file */
  rc = JaycosCreateTransparentEF(ctx, CORPID_JAYCOS_FID_IDENTITY, 128); /* Ciphered ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Create fingerprints file */
  rc = JaycosCreateTransparentEF(ctx, CORPID_JAYCOS_FID_FINGERPRINTS, 170); /* Ciphered ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Create picture file */
  rc = JaycosCreateTransparentEF(ctx, CORPID_JAYCOS_FID_PICTURE, 2464); /* Ciphered ; Key 0 = R/W ; Key 1 = R */
  if (rc)
    return rc;

  /* Create access control file */
  rc = JaycosCreateTransparentEF(ctx, CORPID_JAYCOS_FID_ACCESSCONTROL, 48); /* Plain ; Key 0 = R/W */
  if (rc)
    return rc;

  /*  Create options and signature file */
  rc = JaycosCreateTransparentEF(ctx, CORPID_DESFIRE_FID_OPTIONSSIGN, 32);  /* Plain ; Key 0 = R/W */
  if (rc)
    return rc;

  /* Compute the diversified key 0 */
  rc = ComputeDivKey(ctx, key, CORPID_MASTER_KEY0);
  if (rc)
    return rc;

  /* Put the new key 0 */
  rc = JaycosChangeKey(ctx, CORPID_JAYCOS_KEY0_ID, key);
  if (rc)
    return rc;

  /* Compute the diversified key 1 */
  rc = ComputeDivKey(ctx, key, CORPID_MASTER_KEY1);
  if (rc)
    return rc;

  /* Put the new key 1 */
  rc = JaycosChangeKey(ctx, CORPID_JAYCOS_KEY1_ID, key);
  if (rc)
    return rc;

  /* Done ! */
  return CARDAPPZ_OK;
}
