unit uMfDfIdent;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, ExtCtrls, IniFiles, TJpgImg, jpeg;

type
  TfMfDfIdent = class(TForm)
    status: TStatusBar;
    pTop: TPanel;
    lbTitle: TLabel;
    lbCaption: TLabel;
    imgLogo: TImage;
    pLeft: TPanel;
    img14443: TImage;
    btnReaderOpen: TButton;
    btnReaderClose: TButton;
    lbSnr: TLabel;
    eSnr: TEdit;
    pPicture: TPanel;
    lbPic: TLabel;
    btnPick: TButton;
    lbFirstname: TLabel;
    eFName: TEdit;
    lbLastname: TLabel;
    eLName: TEdit;
    lbBirthdate: TLabel;
    eBDate: TEdit;
    imgLoad: TOpenDialog;
    btnTagWrite: TButton;
    btnTagRead: TButton;
    btnTagFormat: TButton;
    Label1: TLabel;
    eType: TEdit;
    Label2: TLabel;
    eAts: TEdit;
    eUserID: TEdit;
    Label3: TLabel;
    imgHeader: TImage;
    imgPicture: TImage;
    procedure btnReaderOpenClick(Sender: TObject);
    procedure btnReaderCloseClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure btnPickClick(Sender: TObject);
    procedure btnTagReadClick(Sender: TObject);
    procedure btnTagWriteClick(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure btnTagFormatClick(Sender: TObject);
    procedure ClearDisplay(also_data : boolean);
    procedure UpdateStatus;

    function  ReadTagMifare : boolean;
    function  ReadTagDesfire : boolean;
    function  ReadTag : boolean;
    procedure DataToDisplay;

    function  WriteTagDesfire : boolean;
    function  WriteTagMifare : boolean;
    function  WriteTag : boolean;
    function  DisplayToData : boolean;

    function  FormatTagMifare : boolean;
    function  FormatTagDesfire : boolean;
    function  FormatTag : boolean;

    function  SelectTag : boolean;
    procedure LockScreen;
    procedure UnlockScreen;
  private
    picture_filename : string;
    picture_size : dword;
    user_data : array[1..80] of byte;
    picture_data : array[1..3072] of byte;
    reader_active : boolean;
    data_updated : boolean;
  public
    { Dclarations publiques }
  end;

const
   tcl_cid : byte = $FF;

   desfire_key_blank : array[1..16] of byte = (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
   desfire_key_0     : array[1..16] of byte = ($B0,0,0,0,0,0,0,0,0,0,0,0,0,$11,$22,$33);
   desfire_key_1     : array[1..16] of byte = ($A0,0,0,0,0,0,0,0,0,0,0,0,0,$11,$22,$33);

   desfire_aid : dword = $00112233;

   mifare_key_A : array[1..6] of byte = ($A0,$A1,$A2,$A3,$A4,$A5);
   mifare_key_B : array[1..6] of byte = (ord('P'),ord('A'),$00,$11,$22,$33);

   mifare_aid : array[1..16] of byte = ($00, $00, ord('P'),ord('r'),ord('o'),ord('-'),ord('A'),ord('c'),ord('t'),ord('i'),ord('v'),ord('e'), $00, $11, $22, $33);

var
  fMfDfIdent: TfMfDfIdent;

implementation

{$R *.dfm}

uses uSpringProxDll, uSpringProxDsF;

function to_str(p : pointer; l : integer) : string;
var
   i : integer;
   c : ^char;
begin
   result := '';
   c := p;
   for i := 1 to l do
   begin
      if Ord(c^) = 0 then break;
      result := result + c^;
      inc(c);
   end;
end;

function to_hex(p : pointer; l : integer) : string;
var
   i : integer;
   b : ^byte;
begin
   result := '';
   b := p;
   for i := 1 to l do
   begin
      result := result + IntToHex(b^, 2);
      inc(b);
   end;
end;

procedure TfMfDfIdent.DataToDisplay;
var
   f : file of byte;
   d : TDate;
begin
   // Display the information file
   eUserID.Text := to_hex(@user_data[1], 8);
   eLName.Text  := to_str(@user_data[17], 32);
   eFName.Text  := to_str(@user_data[49], 32);
   eBDate.Text := '';
   if ((user_data[15] >= 1) and (user_data[15] <= 12) and (user_data[16] >= 1) and (user_data[16] <= 31)) then 
   try
      d := EncodeDate(1900 + user_data[14], user_data[15], user_data[16]);
      eBDate.Text := DateToStr(d);
   except

   end;

   // Display the picture
   if picture_size > 0 then
   begin
      try
         AssignFile(f, picture_filename);
         Rewrite(f);
      except
         Status.SimpleText := 'Failed to save the picture to temporary file';
         exit;
      end;
      BlockWrite(f, picture_data, picture_size);
      CloseFile(f);

      try
         imgPicture.Picture.LoadFromFile(picture_filename);
         imgPicture.Visible := true;
         Status.SimpleText := 'OK';
      except
         on E : Exception do
         begin
            ShowMessage(E.Message);
            Status.SimpleText := 'Failed to display the picture';
         end;
      end;
   end;

   data_updated := false;
end;

function TfMfDfIdent.DisplayToData : boolean;
var
   f : file of byte;
   d : TDate;
   dy, dm, dd : word;
   i : integer;
   s : string;
begin
   result := false;

   // Format the user buffer
   s := eUserID.Text;
   while length(s) < 16 do s := '0' + s;
   for i := 0 to 7 do
     user_data[i+1] := StrToIntDef('$' + copy(s, 2*i+1, 2), 0);
   eUserID.Text := to_hex(@user_data[1], 8);

   s := eLName.Text;
   if length(s) > 32 then s := copy(s, 1, 32);
   for i := 1 to length(s)    do user_data[16+i] := ord(s[i]);
   for i := length(s)+1 to 32 do user_data[16+i] := 0;

   s := eFName.Text;
   if length(s) > 32 then s := copy(s, 1, 32);
   for i := 1 to length(s)    do user_data[48+i] := ord(s[i]);
   for i := length(s)+1 to 32 do user_data[48+i] := 0;

   try
      d := StrToDate(eBDate.Text);
      DecodeDate(d, dy, dm, dd);
      if ((dy >= 1900) and (dy <= 2050)) then user_data[14] := byte(dy - 1900);
      if ((dm >= 1) and (dm <= 12)) then user_data[15] := byte(dm);
      if ((dd >= 1) and (dd <= 31)) then user_data[16] := byte(dd);
   except

   end;

   // Load picture into buffer
   if imgPicture.Visible then
   begin
      try
         imgPicture.Picture.SaveToFile(picture_filename);
      except
         Status.SimpleText := 'Failed to copy the picture to temporary file';
         exit;
      end;

      try
         AssignFile(f, picture_filename);
         Reset(f);
      except
         Status.SimpleText := 'Failed to load the picture from temporary file';
         exit;
      end;

      picture_size := FileSize(f);
      if picture_size > sizeof(picture_data) then
      begin
         CloseFile(f);
         Status.SimpleText := Format('Picture is too big (%d/%d)', [picture_size, sizeof(picture_data)]);
         exit;
      end;

      BlockRead(f, picture_data, picture_size);
      CloseFile(f);

      for i := picture_size + 1 to sizeof(picture_data) do picture_data[i] := $00;

      user_data[9]  := picture_size div $0100;
      user_data[10] := picture_size mod $0100;
     
      
   end;

   result := true;
end;

procedure TfMfDfIdent.LockScreen;
begin
   Screen.Cursor := crHourglass;
   Application.ProcessMessages;
   Enabled := false;
end;

procedure TfMfDfIdent.UnlockScreen;
begin
   Screen.Cursor := crDefault;
   Application.ProcessMessages;
   Enabled := true;
end;

procedure TfMfDfIdent.UpdateStatus;
begin
   btnReaderOpen.Enabled := not reader_active;
   btnReaderClose.Enabled := reader_active;
   btnTagRead.Enabled := reader_active;
   btnTagFormat.Enabled := reader_active;
   btnTagWrite.Enabled := reader_active;
end;

procedure TfMfDfIdent.btnReaderOpenClick(Sender: TObject);
var
   rc : sword;
   buffer : array[1..256] of char;
begin
   reader_active := false;
   //
   // Try to open the reader
   // ----------------------
   // Setting device to NULL leads to the automatic discovery procedure
   //
   rc := SPROX_ReaderOpen(nil);
   if rc = 0 then
   begin
      // Success

      // Get reader device name
      // ----------------------

      rc := SPROX_ReaderGetDevice(pchar(@buffer), sizeof(buffer));
      if rc = 0 then
      begin
         status.SimpleText := 'Connected, reader found on ' + StrPas(pchar(@buffer));
      end else
      begin
         status.SimpleText := 'Connected, unknown port (' + StrPas(SPROX_GetErrorMessage(rc)) + ')';
      end;

      // OK
      reader_active := true;
   end;
   UpdateStatus;
end;

procedure TfMfDfIdent.btnReaderCloseClick(Sender: TObject);
var
   rc : sword;
begin
   reader_active := false;
   //
   // Disconnect from the reader
   // --------------------------
   //

   rc := SPROX_ReaderClose;
   if rc = 0 then
   begin
      // Success
      Status.SimpleText := 'Connection has been closed';
   end else
   begin
      // Failure
      Status.SimpleText := StrPas(SPROX_GetErrorMessage(rc));
   end;

   ClearDisplay(true);
   UpdateStatus;   
end;

procedure TfMfDfIdent.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   SPROX_ReaderClose;
end;

procedure TfMfDfIdent.FormCreate(Sender: TObject);
var
   ini : TIniFile;
   pic : TPicture;
   str_company : string;
   str_logo    : string;
   str_product : string;
   temp_path   : string;
   t : array[1..1024] of char;
begin
   ini := TIniFile.Create(ExtractFileDir(Application.ExeName) + '\csbinfo.ini');
   str_company := ini.ReadString('company', 'name', 'SpringCard');
   str_logo    := ini.ReadString('company', 'logo', 'logo_pro-active.bmp');
   str_product := ini.ReadString('product', 'name', 'CSB / SpringProx');
   ini.Free;
   pic := TPicture.Create;
   try
      pic.LoadFromFile(str_logo);
      imgLogo.Left := pTop.Width - 8 - pic.Width;
      imgLogo.Picture.Assign(pic);
   except
      // No picture !
   end;
   pic.Free;
   Caption := str_company + ' Corporate identity card Sample Project';
   Application.Title := Caption;

   GetTempPath(1024, pchar(@t));
   temp_path := strpas(pchar(@t));

   picture_filename := temp_path + '\MfDfIdent.jpg';

   data_updated  := false;
   reader_active := false;
end;

procedure TfMfDfIdent.btnPickClick(Sender: TObject);
begin
   if imgLoad.Execute then
   begin
      data_updated := true;
      imgPicture.Visible := false;
      try
         imgPicture.Picture.LoadFromFile(imgLoad.FileName);
         imgPicture.Visible := true;
      except
         Status.SimpleText := 'Failed to open the picture';
      end;
   end;
   UpdateStatus;
end;

procedure TfMfDfIdent.btnTagReadClick(Sender: TObject);
begin
   if data_updated then
   begin
      if MessageDlg('Data have been updated since last read action. Any change made will be lost.',
                    mtInformation,
                    [mbOk, mbCancel],
                    0) <> idOK then exit;
   end;

   LockScreen;
   ClearDisplay(true);

   if not SelectTag then
   begin
      UnlockScreen;
      MessageDlg('Read failed : no compliant tag have been found.',
                 mtInformation,
                 [mbOk],
                 0);
      UpdateStatus;
      exit;
   end;

   if not ReadTag then
   begin
      UnlockScreen;
      MessageDlg('Read failed : tag is not formatted, or has not been written yet.',
                 mtInformation,
                 [mbOk],
                 0);
      UpdateStatus;
      exit;
   end;

   // Display the data
   DataToDisplay;

   Status.SimpleText := 'OK';
   data_updated := false;

   UpdateStatus;
   UnlockScreen;
end;

procedure TfMfDfIdent.btnTagWriteClick(Sender: TObject);
begin
   if (eFName.Text = '') or (eLName.Text = '') then
   begin
      MessageDlg('You must at least enter the name !',
                 mtInformation,
                 [mbOk],
                 0);
      exit;
   end;
   if not imgPicture.Visible then
   begin
      MessageDlg('You must select a picture !',
                 mtInformation,
                 [mbOk],
                 0);
      exit;
   end;

   if not DisplayToData then exit;

   LockScreen;
   ClearDisplay(false);

   if not SelectTag then
   begin
      UnlockScreen;
      MessageDlg('Write failed : no compliant tag have been found.',
                 mtInformation,
                 [mbOk],
                 0);
      UpdateStatus;
      exit;
   end;

   if not WriteTag then
   begin
      UnlockScreen;
      MessageDlg('Write failed : tag is not formatted, or a write error has occured.',
                 mtInformation,
                 [mbOk],
                 0);
      UpdateStatus;
      exit;
   end;

   Status.SimpleText := 'OK';
   data_updated := false;

   UpdateStatus;
   UnlockScreen;
end;

procedure TfMfDfIdent.ClearDisplay(also_data : boolean);
begin
   eType.Text := '';
   eSnr.Text := '';
   eAts.Text := '';
   if also_data then
   begin
      eUserID.Text := '';
      eLName.Text := '';      
      eFName.Text := '';
      eBDate.Text := '';
      imgPicture.Visible := false;
   end;
end;

procedure TfMfDfIdent.DataChange(Sender: TObject);
begin
   data_updated := true;
   UpdateStatus;
end;

procedure TfMfDfIdent.btnTagFormatClick(Sender: TObject);
begin
   if MessageDlg('Formatting the tag will remove ANY existing data on the. Are you sure you want to do this ?',
                 mtWarning,
                 [mbYes, mbNo],
                 0) <> idYes then Exit;

   LockScreen;
   ClearDisplay(false);

   if not SelectTag then
   begin
      UnlockScreen;
      MessageDlg('Format failed : no compliant tag have been found.',
                 mtInformation,
                 [mbOk],
                 0);
      UpdateStatus;
      exit;
   end;

   if not FormatTag then
   begin
      UnlockScreen;
      MessageDlg('Format failed : this application didn''t success in formatting the tag.',
                 mtInformation,
                 [mbOk],
                 0);
      UpdateStatus;
      exit;
   end;

   UpdateStatus;
   UnlockScreen;
end;

function TfMfDfIdent.ReadTag;
var
   rc : boolean;
begin
   rc := false;

   if eType.Text = 'Mif 4K' then
   begin
      rc := ReadTagMifare;
      SPROX_MifStHalt(nil);
   end else
   if eType.Text = 'DesF 4K' then
   begin
      rc := ReadTagDesfire;
      SPROX_TclA_Deselect(tcl_cid);
   end else
   begin
      Status.SimpleText := 'Read cancelled : no compliant tag selected';
   end;

   result := rc;
end;

function TfMfDfIdent.ReadTagMifare : boolean;
var
   rc      : sword;
   s       : byte;
   buffer  : array[1..4096] of byte;
   i, p    : integer;
begin
   result := false;

   for i := 1 to 4096 do buffer[i] := $cc;

   Status.SimpleText := 'Mifare read...';
   Application.ProcessMessages;

   // Read block 1 to retrieve our identifier
   rc := SPROX_MifStReadBlock(nil, 1, @buffer, @mifare_key_a);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to read Mifare block %d (%d)', [1, rc]);
      exit;
   end;

   // Compare block 1 to identifer
   for i := 1 to sizeof(mifare_aid) do
   if buffer[i] <> mifare_aid[i] then
   begin
      Status.SimpleText := 'Mifare application identifier not found';
      exit;
   end;

   // Read tag into big buffer
   p := 1;
   for s := 1 to 39 do
   begin
      rc := SPROX_MifStReadSector(nil, s, @buffer[p], @mifare_key_a);
      if rc <> 0 then
      begin
         Status.SimpleText := Format('Failed to read Mifare sector %d (%d)', [s, rc]);
         exit;
      end;

      if s < 32 then p := p + 48 else p := p + 240;
   end;

   // The first 80 bytes of the big buffer are the user part
   for i := 1 to 80 do
      user_data[i] := buffer[i];

   // Retrieve size of picture from data
   picture_size := user_data[9] * $0100 + user_data[10];
   if (picture_size > sizeof(picture_data)) then
   begin
      Status.SimpleText := Format('Mifare picture has an invalid size (%d)', [picture_size]);
      picture_size := 0;
   end;

   // Picture starts at offset 528 in tag, up to length specified in user part
   for i := 1 to picture_size do
      picture_data[i] := buffer[528-48+i];
   for i := picture_size + 1 to sizeof(picture_data) do
      picture_data[i] := $BC;

   result := true;
end;

function TfMfDfIdent.ReadTagDesfire : boolean;
var
   rc : sword;
   i : dword;
   buffer : array[1..4096] of byte;
   size : dword;
begin
   result := false;

   Status.SimpleText := 'Reading Desfire tag...';
   Application.ProcessMessages;

   // Select our application
   rc := DesFireCard_SelectApplication(desfire_aid);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Desfire application not found';
      exit;
   end;

   // Get authenticated with key #1 for reading
   rc  := DesFireCard_Authenticate(1, @desfire_key_1);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Authentication failed';
      exit;
   end;

   Status.SimpleText := 'Application selected...';
   Application.ProcessMessages;

   // The first file stores the text informations
   size := sizeof(buffer);
   rc := DesFireCard_ReadData(0, 0, 0, 0, @buffer, @size);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Desfire information file read error (%d)', [rc]);
      exit;
   end;
   if size <> 80 then
   begin
      Status.SimpleText := Format('Desfire information file has an invalid size (%d)', [size]);
      exit;
   end;
   for i := 1 to 80 do
      user_data[i] := buffer[i];

   // Retrieve size of picture from data
   picture_size := user_data[9] * $0100 + user_data[10];
   if (picture_size > sizeof(picture_data)) then
   begin
      Status.SimpleText := Format('Desfire picture has an invalid size (%d)', [picture_size]);
      picture_size := 0;
   end;

   if picture_size > 0 then
   begin
      // The second file stores the picture
      size := sizeof(buffer);
      rc := DesFireCard_ReadData(1, 0, 0, 0, @buffer, @size);
      if rc <> 0 then
      begin
         Status.SimpleText := Format('Desfire picture file read error (%d)', [rc]);
         exit;
      end;

      // Truncate to actual length
      if size < picture_size then picture_size := size;

      // Copy data
      for i := 1 to picture_size do
         picture_data[i] := buffer[i];
   end;

   for i := picture_size + 1 to sizeof(picture_data) do
      picture_data[i] := $BC;

   result := true;
end;

function TfMfDfIdent.FormatTagMifare : boolean;
var
   rc     : sword;
   s      : byte;
   buffer : array[1..240] of byte;
   i      : integer;
begin
   result := false;

   Status.SimpleText := 'Formatting Mifare tag...';
   Application.ProcessMessages;

   for i := 1 to sizeof(buffer) do buffer[i] := $00;

   for s := 0 to 39 do
   begin
      rc := SPROX_MifStUpdateAccessBlock(nil, s, @mifare_key_B, @mifare_key_A, @mifare_key_B, acc_normal, acc_normal, acc_normal, acc_security);
      if rc <> 0 then
      rc := SPROX_MifStUpdateAccessBlock(nil, s, nil, @mifare_key_A, @mifare_key_B, acc_normal, acc_normal, acc_normal, acc_security);
      if rc <> 0 then
      begin
         Status.SimpleText := Format('Failed to format Mifare sector %d (%d)', [s, rc]);
         exit;
      end;

      if s = 0 then
      begin
         // On first sector we write our identifier
         rc := SPROX_MifStWriteBlock(nil, 1, @mifare_aid, @mifare_key_b);
         if rc = 0 then
         rc := SPROX_MifStWriteBlock(nil, 2, @buffer, @mifare_key_b);
      end else
      begin
         rc := SPROX_MifStWriteSector(nil, s, @buffer, @mifare_key_b);
      end;

      if rc <> 0 then
      begin
         Status.SimpleText := Format('Failed to initialize Mifare sector %d (%d)', [s, rc]);
         exit;
      end;
   end;

   result := true;
end;

function TfMfDfIdent.FormatTagDesfire : boolean;
var
   rc : sword;
begin
   result := false;

   Status.SimpleText := 'Formatting Desfire tag...';
   Application.ProcessMessages;

   // Select root application
   rc := DesFireCard_SelectApplication(0);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to select Desfire root application';
      exit;
   end;

   // Try to get authenticated with blank key as master key
   rc := DesFireCard_Authenticate(0, @desfire_key_blank);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to authenticate with Desfire master key (not blank ?)';
      exit;
   end;

   // Format the Desfire card
   rc := DesFireCard_FormatPicc;
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to format the Desfire card';
      exit;
   end;

   // Select root application again
   rc := DesFireCard_SelectApplication(0);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to select Desfire root application after format';
      exit;
   end;

   // Now card is blank, let's authenticate with blank key as master key
   rc := DesFireCard_Authenticate(0, @desfire_key_blank);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to authenticate on the blank Desfire card';
      exit;
   end;

   // We can now create our own application on the card
   rc := DesFireCard_CreateApplication(desfire_aid, $e3, 2);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to create the Desfire application';
      exit;
   end;

   // Select our application
   rc := DesFireCard_SelectApplication(desfire_aid);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Failed to select the Desfire application';
      exit;
   end;

   // Get authenticated with blank key 0
   rc  := DesFireCard_Authenticate(0, @desfire_key_blank);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Authentication failed';
      exit;
   end;

   // Create information file
   rc := DesFireCard_CreateStdDataFile(0, 0, $1f00, sizeof(user_data));
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to create the information file (%d)', [rc]);
      exit;
   end;

   // Create picture file
   rc := DesFireCard_CreateStdDataFile(1, 0, $1f00, sizeof(picture_data));
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to create the picture file (%d)', [rc]);
      exit;
   end;

   // Create fingerprint file
   rc := DesFireCard_CreateStdDataFile(2, 3, $1f00, 340);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to create the fingerprints file (%d)', [rc]);
      exit;
   end;

   // Create balance
   rc := DesFireCard_CreateValueFile(3, 1, $1f00, 0, 100000, 0, false);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to create the balance file (%d)', [rc]);
      exit;
   end;

   // Create history
   rc := DesFireCard_CreateCyclicRecordFile(4, 0, $eeee, 8, 7);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to create the history file (%d)', [rc]);
      exit;
   end;

   // Put the new keys
   rc := DesFireCard_ChangeKey(0, @desfire_key_0, nil);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Change key failed (key 0)';
      exit;
   end;

   rc  := DesFireCard_Authenticate(1, @desfire_key_blank);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Authentication failed (key 1)';
      exit;
   end;

   rc := DesFireCard_ChangeKey(1, @desfire_key_1, nil);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Change key failed (key 1)';
      exit;
   end;

   result := true;
end;

function TfMfDfIdent.FormatTag : boolean;
var
   rc : boolean;
begin
   rc := false;

   if eType.Text = 'Mif 4K' then
   begin
      rc := FormatTagMifare;
   end else
   if eType.Text = 'DesF 4K' then
   begin
      rc := FormatTagDesfire;
      // T=CL deselect
      SPROX_TclA_Deselect(tcl_cid);
   end else
   begin
      Status.SimpleText := 'Format cancelled : no compliant tag selected';
   end;

   // Halt the tag anyway
   SPROX_TclA_Halt();

   if rc then Status.SimpleText := 'OK';

   result := rc;
end;

function TfMfDfIdent.WriteTagDesfire : boolean;
var
   rc : sword;
begin
   result := false;

   Status.SimpleText := 'Writing Desfire tag...';
   Application.ProcessMessages;

   // Select our application
   rc := DesFireCard_SelectApplication(desfire_aid);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Desfire application not found';
      exit;
   end;

   // Get authenticated with key #0 for writing
   rc  := DesFireCard_Authenticate(0, @desfire_key_0);
   if rc <> 0 then
   begin
      Status.SimpleText := 'Authentication failed';
      exit;
   end;

   // Write the data
   rc := DesFireCard_WriteData2(0, 0, sizeof(user_data), @user_data);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to write the information file (%d)', [rc]);
      exit;
   end;

   Status.SimpleText := 'Information file written...';
   Application.ProcessMessages;

   // Write the picture
   if picture_size > 0  then
   begin

      rc := DesFireCard_WriteData2(1, 0, picture_size, @picture_data);
      if rc <> 0 then
      begin
         Status.SimpleText := Format('Failed to write the picture file (%d)', [rc]);
         exit;
      end;

   end;

   result := true;
end;

function TfMfDfIdent.WriteTagMifare : boolean;
var
   rc : sword;
   i : dword;
   p : integer;
   b, s : byte;
   buffer : array[1..32] of byte;
begin
   result := false;

   Status.SimpleText := 'Writing Mifare tag...';
   Application.ProcessMessages;

   // Read block 1 to retrieve our identifier
   rc := SPROX_MifStReadBlock(nil, 1, @buffer, @mifare_key_a);
   if rc <> 0 then
   begin
      Status.SimpleText := Format('Failed to read Mifare block %d (%d)', [1, rc]);
      exit;
   end;

   // Compare block 1 to identifer
   for i := 1 to sizeof(mifare_aid) do
   if buffer[i] <> mifare_aid[i] then
   begin
      Status.SimpleText := 'Mifare application identifier not found';
      exit;
   end;

   // Write user information block
   p := 1;

   for b := 4 to 9 do
   begin
      if b <> 7 then // Skip security block
      begin
         rc := SPROX_MifStWriteBlock(nil, b, @user_data[p], @mifare_key_b);
         if rc <> 0 then
         begin
            Status.SimpleText := Format('Failed to write Mifare block %d (%d)', [b, rc]);
            exit;
         end;
         p := p + 16;
      end;
   end;

   // Now write the picture, starting at sector 11
   if picture_size > 0  then
   begin

      p := 1;
      for s := 11 to 39 do
      begin

         rc := SPROX_MifStWriteSector(nil, s, @picture_data[p], @mifare_key_b);
         if rc <> 0 then
         begin
            Status.SimpleText := Format('Failed to write Mifare sector %d (%d)', [s, rc]);
            exit;
         end;

         // Next offset depends on sector size
         if s < 32 then p := p + 48 else p := p + 240;

      end;
   end;


   result := true;
end;

function TfMfDfIdent.WriteTag : boolean;
var
   rc : boolean;
begin
   rc := false;

   if eType.Text = 'Mif 4K' then
   begin
      rc := WriteTagMifare;
   end else
   if eType.Text = 'DesF 4K' then
   begin
      rc := WriteTagDesfire;
      // T=CL deselect
      SPROX_TclA_Deselect(tcl_cid);
   end else
   begin
      Status.SimpleText := 'Write cancelled : no compliant tag selected';
   end;

   // Halt the tag anyway
   SPROX_TclA_Halt();

   result := rc;
end;

function TfMfDfIdent.SelectTag : boolean;
var
   str   : string;
   rc : sword;
   i, j  : integer;
   snr    : array[1..12] of byte;
   atq    : array[1..2] of byte;
   atq_w  : word;
   sak    : array[1..1] of byte;
   sak_b  : byte;
   ats    : array[1..32] of byte;
   atslen : byte;
begin
   result := false;

   Status.SimpleText := 'Looking for a tag...';
   Application.ProcessMessages;

   // Activate another (or the same...) tag
   rc := SPROX_MifStSelectAny(@snr, @atq, @sak);
   if rc <> 0 then
   begin
      Status.SimpleText := 'No tag on the reader';
      Exit;
   end;

   // Display serial number. Length depend on atq LSB
   j := 0;
   if (atq[1] and $C0 = $00) then j := 4  else  // simple
   if (atq[1] and $C0 = $40) then j := 7  else  // double
   if (atq[1] and $C0 = $40) then j := 10 else  // triple
   if (atq[1] and $C0 = $C0) then j := 4;       // RFU

   str := '';
   for i := 1 to j do str := str + IntToHex(snr[i], 2);
   eSnr.Text := str;
   Application.ProcessMessages;

   // Try to recognize the card
   atq_w := atq[1] + atq[2] * $0100;
   sak_b := sak[1];

   if (atq_w = $0010) and (sak_b = $81) then
   begin
      eType.Text := 'Mif Lt';
      SPROX_MifStHalt(nil);
      exit;
   end;

   if (atq_w = $0044) and (sak_b = $00) then
   begin
      eType.Text := 'Mif UL';
      SPROX_MifStHalt(nil);
      exit;
   end;

   if (atq_w = $0004) then
   begin
      eType.Text := 'Mif 1K';
      SPROX_MifStHalt(nil);
      exit;
   end;

   if (atq_w = $0002) then
   begin
      eType.Text := 'Mif 4K';
      Status.SimpleText := 'Mifare 4k tag selected...';
      result := true;
      exit;
   end;

   if (atq_w = $0344) and (sak_b = $20) then
   begin
      eType.Text := 'DesF 4K';

      atslen := sizeof(ats);
      rc := SPROX_TclA_GetAts(tcl_cid, @ats, @atslen);
      if rc <> 0 then
      begin
         Status.SimpleText := StrPas(SPROX_GetErrorMessage(rc));
         exit;
      end;

      str := '';
      for i := 1 to atslen do str := str + IntToHex(ats[i], 2);
      eAts.Text := str;

      Status.SimpleText := 'Desfire tag selected...';
      result := true;
      exit;
   end;

   eType.Text := '???';
end;

end.
