#pragma hdrstop

#include "uSettings.h"
#include <stdio.h>

TSettings *Settings;

TSettings::TSettings()
{
  //
}

bool TSettings::SaveSettings()
{
  bool rc = false;
  AnsiString CommandLine;

  TRegistry *Registry = new TRegistry(KEY_WRITE);

  CommandLine = ExtractFilePath(Application->ExeName);
  CommandLine = IncludeTrailingPathDelimiter(CommandLine);
  CommandLine = CommandLine + "ScanUid.exe";

  try
  {
    Registry->RootKey = HKEY_LOCAL_MACHINE;
    Registry->CreateKey("Software");
    Registry->CreateKey("Software\\Pro-Active");
    Registry->OpenKey("Software\\Pro-Active\\ScanUid", true);

    Registry->WriteString("Device", Device);
    Registry->WriteString("ScanMode", IntToStr(ScanMode));

    Registry->WriteString("Prefix", Prefix);
    Registry->WriteString("Postfix", Postfix);
    Registry->WriteString("ScanDelay", IntToStr(ScanDelay));
    Registry->WriteString("ScanFilter", IntToStr(ScanFilter));
    Registry->WriteString("ScanBeep", ScanBeep ? "1" : "0");
    Registry->WriteString("RemoveLeading0", RemoveLeading0 ? "1" : "0");

    Registry->WriteString("MifareBlock", IntToStr(MifareBlock));
    Registry->WriteString("MifareOffset", IntToStr(MifareOffset));
    Registry->WriteString("MifareSize", IntToStr(MifareSize));
    Registry->WriteString("MifareKey", MifareKey);
    Registry->WriteString("MifareFormat", IntToStr(MifareFormat));
    Registry->WriteString("MifareRevert", MifareRevert ? "1" : "0");

    Registry->WriteString("14443ASize", IntToStr(Iso14443aSize));
    Registry->WriteString("14443AFormat", IntToStr(Iso14443aFormat));
    Registry->WriteString("14443ARevert", Iso14443aRevert ? "1" : "0");
    Registry->WriteString("14443APrefixAtq", Iso14443aATQ ? "1" : "0");
    Registry->WriteString("14443APostfixSak", Iso14443aSAK ? "1" : "0");
    Registry->WriteString("14443APrefixA", Iso14443aA ? "1" : "0");

    Registry->WriteString("14443BSize", IntToStr(Iso14443bSize));
    Registry->WriteString("14443BFormat", IntToStr(Iso14443bFormat));
    Registry->WriteString("14443BRevert", Iso14443bRevert ? "1" : "0");
    Registry->WriteString("14443BPrefixA", Iso14443bB ? "1" : "0");

    rc = true;
  }
  __finally
  {
    delete Registry;
  }

  Registry = new TRegistry(KEY_WRITE);
  try
  {
    Registry->RootKey = HKEY_LOCAL_MACHINE;
    Registry->OpenKey("Software\\Microsoft\\Windows\\CurrentVersion\\Run", false);

    if (StartMode & 0x02)
    {
      Registry->WriteString("SPROX_ScanUid", CommandLine);
    } else
    {
      Registry->DeleteValue("SPROX_ScanUid");
    }
  }
  __finally
  {
    delete Registry;
  }

  Registry = new TRegistry(KEY_WRITE);
  try
  {
    Registry->OpenKey("Software\\Microsoft\\Windows\\CurrentVersion\\Run", false);

    if (StartMode & 0x01)
    {
      Registry->WriteString("SPROX_ScanUid", CommandLine);
    } else
    {
      Registry->DeleteValue("SPROX_ScanUid");
    }
  }
  __finally
  {
    delete Registry;
  }

  return rc;
}

void TSettings::LoadSettings()
{
  TRegistry *Registry = new TRegistry(KEY_READ);

  try
  {
    Registry->RootKey = HKEY_LOCAL_MACHINE;
    Registry->OpenKey("Software\\Pro-Active\\ScanUid", false);

    Device = Registry->ReadString("Device");
    ScanMode = StrToIntDef(Registry->ReadString("ScanMode"), 0);

    Prefix = Registry->ReadString("Prefix");
    Postfix = Registry->ReadString("Postfix");
    ScanDelay = StrToIntDef(Registry->ReadString("ScanDelay"), 1);
    ScanFilter = StrToIntDef(Registry->ReadString("ScanFilter"), -1);
    ScanBeep =
      StrToIntDef(Registry->ReadString("ScanBeep"), 0) ? true : false;
    RemoveLeading0 =
      StrToIntDef(Registry->ReadString("RemoveLeading0"), 0) ? true : false;

    MifareBlock = StrToIntDef(Registry->ReadString("MifareBlock"), 4);
    if (MifareBlock > 64)
      MifareBlock = 4;
    MifareOffset = StrToIntDef(Registry->ReadString("MifareOffset"), 0);
    if (MifareOffset > 15)
      MifareOffset = 0;
    MifareSize = StrToIntDef(Registry->ReadString("MifareSize"), 4);
    if (MifareSize == 0)
      MifareSize = 4;
    if (MifareSize > 16)
      MifareSize = 4;
    MifareKey = Registry->ReadString("MifareKey");
    MifareFormat = StrToIntDef(Registry->ReadString("MifareFormat"), 0);
    MifareRevert =
      StrToIntDef(Registry->ReadString("MifareRevert"), 0) ? true : false;

    Iso14443aSize = StrToIntDef(Registry->ReadString("14443ASize"), 1);
    Iso14443aFormat = StrToIntDef(Registry->ReadString("14443AFormat"), 0);
    Iso14443aRevert =
      StrToIntDef(Registry->ReadString("14443ARevert"), 1) ? true : false;
    Iso14443aATQ =
      StrToIntDef(Registry->ReadString("14443APrefixAtq"), 0) ? true : false;
    Iso14443aSAK =
      StrToIntDef(Registry->ReadString("14443APostfixSak"), 0) ? true : false;
    Iso14443aA =
      StrToIntDef(Registry->ReadString("14443APrefixA"), 0) ? true : false;

    Iso14443bSize = StrToIntDef(Registry->ReadString("14443BSize"), 1);
    Iso14443bFormat = StrToIntDef(Registry->ReadString("14443BFormat"), 0);
    Iso14443bRevert =
      StrToIntDef(Registry->ReadString("14443BRevert"), 0) ? true : false;
    Iso14443bB =
      StrToIntDef(Registry->ReadString("14443BPrefixB"), 0) ? true : false;

  }
  __finally
  {
    delete Registry;
  }

  StartMode = 0;

  Registry = new TRegistry(KEY_READ);
  try
  {
    Registry->RootKey = HKEY_LOCAL_MACHINE;
    Registry->OpenKey("Software\\Microsoft\\Windows\\CurrentVersion\\Run", false);
    if (Registry->ReadString("SPROX_ScanUid") != "")
    {
      StartMode |= 0x02;
    }
  }
  __finally
  {
    delete Registry;
  }

  Registry = new TRegistry(KEY_READ);
  try
  {
    Registry->RootKey = HKEY_CURRENT_USER;
    Registry->OpenKey("Software\\Microsoft\\Windows\\CurrentVersion\\Run", false);
    if (Registry->ReadString("SPROX_ScanUid") != "")
    {
      StartMode |= 0x01;
    }
  }
  __finally
  {
    delete Registry;
  }
}

void TSettings::BroadcastSettingsUpdated()
{
  //
}


AnsiString TSettings::Format14443B(unsigned char atq[11])
{
  AnsiString result = "";
  char tmp[16 + 1];

  if (Iso14443bFormat == 1)
  {
    unsigned long dw;

    if (Iso14443bRevert)
    {
      dw = atq[3];
      dw <<= 8;
      dw += atq[2];
      dw <<= 8;
      dw += atq[1];
      dw <<= 8;
      dw += atq[0];
    } else
    {
      dw = atq[0];
      dw <<= 8;
      dw += atq[1];
      dw <<= 8;
      dw += atq[2];
      dw <<= 8;
      dw += atq[3];
    }

    sprintf(tmp, "%010lu", dw);
    result = tmp;

  } else
  {
    int i, l;

    if (Iso14443bSize == 0)
      l = 11;
    else
      l = 4;

    for (i = 0; i < l; i++)
    {
      if (Iso14443bRevert)
      {
        sprintf(tmp, "%02X", atq[l - 1 - i]);
      } else
      {
        sprintf(tmp, "%02X", atq[i]);
      }
      result = result + tmp;
    }
  }

  if (Iso14443bB)
  {
    result = "B" + result;
  }

  if (RemoveLeading0)
  {
    while (result.Pos("0") == 1)
      result.Delete(1, 1);
  }

  return result;
}

AnsiString TSettings::Format14443A(unsigned char atq[2],
                                   unsigned char uid[12],
                                   unsigned char uidlen, unsigned char sak[1])
{
  AnsiString result = "";
  char tmp[16 + 1];

  if (Iso14443aFormat == 1)
  {
    unsigned long dw;

    if (Iso14443aRevert)
    {
      dw = uid[3];
      dw <<= 8;
      dw += uid[2];
      dw <<= 8;
      dw += uid[1];
      dw <<= 8;
      dw += uid[0];
    } else
    {
      dw = uid[0];
      dw <<= 8;
      dw += uid[1];
      dw <<= 8;
      dw += uid[2];
      dw <<= 8;
      dw += uid[3];
    }

    sprintf(tmp, "%010lu", dw);
    result = tmp;

  } else
  {
    int i, l;

    if (Iso14443aSize == 0)
      l = uidlen;
    else
      l = 4;

    for (i = 0; i < l; i++)
    {
      if (Iso14443aRevert)
      {
        sprintf(tmp, "%02X", uid[l - 1 - i]);
      } else
      {
        sprintf(tmp, "%02X", uid[i]);
      }
      result = result + tmp;
    }
  }

  if (Iso14443aATQ)
  {
    sprintf(tmp, "%02X%02X", atq[1], atq[0]);
    result = tmp + result;
  }

  if (Iso14443aSAK)
  {
    sprintf(tmp, "%02X", sak[0]);
    result = result + tmp;
  }

  if (Iso14443aA)
  {
    result = "A" + result;
  }

  if (RemoveLeading0)
  {
    while (result.Pos("0") == 1)
      result.Delete(1, 1);
  }

  return result;
}

AnsiString TSettings::FormatMifare(unsigned char data[16])
{
  AnsiString result = "";
  char tmp[16 + 1];
  int i, l;

  if (MifareFormat == 2)
  {
    for (i = 0; i < MifareSize; i++)
    {
      sprintf(tmp, "%c", data[MifareOffset + i]);
      if (!strlen(tmp))
        break;
      if (MifareRevert)
        result = tmp + result;
      else
        result = result + tmp;
    }
  } else if (MifareFormat == 1)
  {
    unsigned long dw;

    if (MifareRevert)
    {
      dw = data[MifareOffset + 3];
      dw <<= 8;
      dw += data[MifareOffset + 2];
      dw <<= 8;
      dw += data[MifareOffset + 1];
      dw <<= 8;
      dw += data[MifareOffset + 0];
    } else
    {
      dw = data[MifareOffset + 0];
      dw <<= 8;
      dw += data[MifareOffset + 1];
      dw <<= 8;
      dw += data[MifareOffset + 2];
      dw <<= 8;
      dw += data[MifareOffset + 3];
    }

    sprintf(tmp, "%010lu", dw);
    result = tmp;

  } else
  {
    for (i = 0; i < MifareSize; i++)
    {
      sprintf(tmp, "%02X", data[MifareOffset + i]);
      if (MifareRevert)
        result = tmp + result;
      else
        result = result + tmp;
    }
  }

  if (RemoveLeading0)
  {
    while (result.Pos("0") == 1)
      result.Delete(1, 1);
  }

  return result;
}

bool TSettings::GetMifareKey(unsigned char dst[6])
{
  int i;
  char *src;

  if (dst == NULL)
    return false;

  src = MifareKey.c_str();
  if (src == NULL)
    return false;

  if (strlen(src) == 6)
  {
    for (i = 0; i < 6; i++)
      dst[i] = src[i];
  } else if (strlen(src) == 12)
  {
    for (i = 0; i < 12; i++)
    {
      unsigned char b;

      if ((src[i] >= '0') && (src[i] <= '9'))
        b = src[i] - '0';
      else if ((src[i] >= 'A') && (src[i] <= 'F'))
        b = src[i] - 'A' + 10;
      else if ((src[i] >= 'a') && (src[i] <= 'f'))
        b = src[i] - 'a' + 10;
      else
        return false;

      if (i % 2)
        dst[i / 2] |= b;
      else
        dst[i / 2] = b << 4;
    }

  } else
    return false;

  return true;
}
